

class CtypedException(Exception):
    """"""


class FunctionRedeclared(CtypedException):
    """"""


class UnsupportedTypeError(CtypedException):
    """"""


class TypehintError(CtypedException):
    """"""
